// DCC Decoder for DS-DCC decode
// By yaasan
// Based on Nicolas's sketch http://blog.nicolas.cx
// Inspired by Geoff Bunza and his 17 Function DCC Decoder & updated library
//
//
// Debug serial output available on the serial port at baud 115200, aka Tools -> Serial Monitor
//

//#include <NmraDcc.h>
#include "NmraDcc_diet.h"
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below

//#define _DEBUG
//#define DSFOX // DSFox Debug board
#define A4953 // set pwm to On_Brake mode

//各種設定、宣言
#define VERSIONID    10  // ver 0.1

#define DECODER_ADDRESS 3
#define DCC_ACK_PIN 9	//if defined enables the ACK pin functionality. Comment out to disable.

#ifdef DSFOX
 #define MOTOR_PWM_A  3
 #define MOTOR_PWM_B 11
 #define PIN_FHEAD   A3
 #define PIN_RHEAD   A4
 #define RUN_LED      5
#else
 #if defined(__AVR_ATtiny85__)
   #define MOTOR_PWM_A  1
   #define MOTOR_PWM_B  4
 #else
   #define MOTOR_PWM_A  9
   #define MOTOR_PWM_B 10
 #endif
 #define PIN_FHEAD    0
 #define PIN_RHEAD    3
#endif

//#define MOTOR_LIM_MAX 255

#define CV_VSTART     2
#define CV_ACCRATIO   3
#define CV_DECCRATIO  4
#define CV_VMAX       5
#define CV_CONSIST   19
#define CV_BRIGHT    33

//使用クラスの宣言
NmraDcc	 Dcc;
DCC_MSG	 Packet;

//Task Schedule
unsigned long gPreviousL1 = 0;
unsigned long gPreviousL5 = 0;

//モータ制御関連の変数
uint32_t gPwmLPF_buf = 0;
uint32_t gPwmLPF2_buf = 0;
uint8_t gPwmDirv = 128;
uint16_t gSpeedCmd = 0;
uint16_t gPrevSpeed = 0;

//http://www.technoblogy.com/show?LE0
#if defined(__AVR_ATtiny85__)
const int motor_pwma = 0;
const int motor_pwmb = 1;
volatile uint8_t* Port[] = {&OCR1A, &OCR1B};
#endif

//ファンクションの変数
uint8_t gFuncBits[5] = {0};

//CV
uint8_t gCV19_Consist = 0;
uint8_t gConsistDir = 0;
uint8_t gCV29_Conf = 0;
uint8_t gCV29Dir = 0;

uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;
uint8_t gCV2_Vstart = 20;
uint8_t gCV3_AccRatio = 16;
uint8_t gCV4_DecRatio = 16;
uint8_t gCV5_VMAX = 255;
uint8_t gCV33_Bright = 0;
uint8_t gBrightCount = 0;
uint8_t gRunCount = 0;

//Internal variables and other.
#if defined(DCC_ACK_PIN)
const int DccAckPin = DCC_ACK_PIN ;
#endif

boolean gSvcMode = false;

struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};
CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 2 },	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
  {CV_VSTART, 20},
  {CV_ACCRATIO, 16},
  {CV_DECCRATIO, 16},
  {CV_VMAX, 255},
  {CV_CONSIST, 0},
  {CV_BRIGHT, 0},
};

void(* resetFunc) (void) = 0;  //declare reset function at address 0


uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);

#if defined(__AVR_ATtiny85__)
void pwmWrite(int dir, int intensity);
#endif

uint16_t limitSpeed(uint16_t inSpeed);

void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

	resetCVToDefault();	 
	
	#ifdef _DEBUG
	Serial.println("Resetting...");
	#endif
	
	delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying

	resetFunc();
};


void resetCVToDefault()
{
	//CVをデフォルトにリセット
	#ifdef _DEBUG
	Serial.println("CVs being reset to factory defaults");
	#endif
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
};

extern void notifyServiceMode(bool svc)
{
   gSvcMode = svc;
}

extern void	   notifyCVChange( uint16_t CV, uint8_t Value)
{
  if(!gSvcMode)
  {
    switch(CV) {
      case CV_VSTART:
        gCV2_Vstart = Value;
        Dcc.setCV(CV, gCV2_Vstart);
        break;
      case CV_ACCRATIO:
        gCV3_AccRatio = Value;
        Dcc.setCV(CV, gCV3_AccRatio);
        break;
      case CV_DECCRATIO:
        gCV4_DecRatio = Value;
        Dcc.setCV(CV, gCV4_DecRatio);
        break;
      case CV_VMAX:
        gCV5_VMAX = Value;
        Dcc.setCV(CV, gCV5_VMAX);
        break;
      case CV_BRIGHT:
        gCV33_Bright = Value;
        Dcc.setCV(CV, gCV33_Bright);
        break;
        
      case CV_29_CONFIG:
        gCV29_Conf = Value;
        Dcc.setCV(CV, gCV29_Conf);
        if ((gCV29_Conf & 0x01) > 0)
          gCV29Dir = 1;
        else
          gCV29Dir = 0;
        break;
    }
  }

   //CVが変更されたときのメッセージ
   #ifdef _DEBUG
   Serial.print("CV "); 
   Serial.print(CV); 
   Serial.print(" Changed to "); 
   Serial.println(Value, DEC);
   #endif
};

void notifyCVAck(void)
{
  #ifdef _DEBUG
  Serial.println("notifyCVAck");
  #endif
  
  if(gSvcMode)
  {
    #if defined(__AVR_ATtiny85__)
    pwmWrite(motor_pwma, 200);
    pwmWrite(motor_pwmb, 0);
    #else
    analogWrite(MOTOR_PWM_A, 200);
    analogWrite(MOTOR_PWM_B, 0);
    #endif
    
    delay( 6 );  
    
    #if defined(__AVR_ATtiny85__)
    pwmWrite(motor_pwma, 0);
    #else
    analogWrite(MOTOR_PWM_A, 0);
    #endif
  }
}


void setup()
{
	uint8_t cv_value;
	
	//シリアル通信開始
	#ifdef _DEBUG
	Serial.begin(115200);
	#endif

	//PWM キャリア周期:31kHz
	#if defined(__AVR_ATtiny85__)
	  TCCR1 = 1<<PWM1A | 2<<COM1A0 | 1<<CS10;
	  GTCCR = 1<<PWM1B | 2<<COM1B0;
	#else
	  #ifdef DSFOX
	  // OC2A, OC2B
	  TCCR2B &= B11111000;
	  TCCR2B |= B00000001;
	  #else
	  // OC1A, OC1B
	  TCCR1B &= B11111000;
	  TCCR1B |= B00000001;
	  #endif
	#endif
	
	//PWMを10bit化
	//TCCR1A |= B00000011;
	
	//PWM出力ピンを出力にセット
	pinMode(MOTOR_PWM_A, OUTPUT);
	pinMode(MOTOR_PWM_B, OUTPUT);
	
	//ファンクションの割り当てピン初期化
	pinMode(PIN_FHEAD, OUTPUT);
	digitalWrite(PIN_FHEAD, 1);
	pinMode(PIN_RHEAD, OUTPUT);
	digitalWrite(PIN_RHEAD, 1);
	#ifdef DSFOX
	pinMode(RUN_LED, OUTPUT);
	digitalWrite(RUN_LED, 1);
	#endif
	
	//DCCの応答用負荷ピン
	
	#if defined(DCCACKPIN)
	//Setup ACK Pin
	//pinMode(DccAckPin,OUTPUT);
	//digitalWrite(DccAckPin, 0);
	#endif
  
 

  #if !defined(DECODER_DONT_DEFAULT_CV_ON_POWERUP)
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF ){	 //if eeprom has 0xFF then assume it needs to be programmed
	  #ifdef _DEBUG
	  Serial.println("CV Defaulting due to blank eeprom");
	  #endif
	  
	  notifyCVResetFactoryDefault();
	  
   } else{
	#ifdef _DEBUG
	 Serial.println("CV Not Defaulting");
	 #endif
   }
  #else
	 #ifdef _DEBUG
	 Serial.println("CV Defaulting Always On Powerup");
	 #endif
	 
	 notifyCVResetFactoryDefault();
  #endif 
  
   
   
  // Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
  Dcc.pin(0, 2, 0);
  
  // Call the main DCC Init function to enable the DCC Receiver
  //Dcc.init( MAN_ID_DIY, VERSIONID, FLAGS_MY_ADDRESS_ONLY , 0 ); 
  Dcc.init( MAN_ID_DIY, VERSIONID, FLAGS_OUTPUT_ADDRESS_MODE, 0 );
  
  //Init CVs
  gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
  gCV2_Vstart = Dcc.getCV( CV_VSTART ) ;
  gCV3_AccRatio = Dcc.getCV( CV_ACCRATIO ) ;
  gCV4_DecRatio = Dcc.getCV( CV_DECCRATIO ) ;
  gCV5_VMAX = Dcc.getCV( CV_VMAX ) ;
  gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8)
              + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
  gCV19_Consist = Dcc.getCV(CV_CONSIST);
  gConsistDir = gCV19_Consist & 0b10000000;
  gCV19_Consist = gCV19_Consist & 0b01111111;
  gCV33_Bright = Dcc.getCV(CV_BRIGHT);
  
  gCV29_Conf = Dcc.getCV( CV_29_CONFIG ) ;
  //cv29 Direction Check
  if ((gCV29_Conf & 0x01) > 0) {
    gCV29Dir = 1;//REVをFWDにする
  } else {
    gCV29Dir = 0;//FWDをFWDにする
  }

  #ifdef _DEBUG
  Serial.print("CV1(ShortAddr): ");
  Serial.println(gCV1_SAddr);
  Serial.print("CV17/18(LongAddr): ");
  Serial.println(gCVx_LAddr);
  Serial.print("CV2(Vstart): ");
  Serial.println(gCV2_Vstart);
  Serial.print("CV3(AccRatio): ");
  Serial.println(gCV3_AccRatio);
  Serial.print("CV4(DecRatio): ");
  Serial.println(gCV4_DecRatio);
  Serial.print("CV5(MaxSpeed): ");
  Serial.println(gCV5_VMAX);
  Serial.print("CV19: ");
  Serial.println(gCV19_Consist);
  Serial.print("ConsistDir: ");
  Serial.println(gConsistDir);
  Serial.print("CV33(Bright): ");
  Serial.println(gCV33_Bright);
  Serial.print("CV29: ");
  Serial.println(gCV29_Conf);
  
  Serial.println("Ready");
  #endif
  
  //Reset task
  gPreviousL1 = millis();
  gPreviousL5 = millis();
}

void loop()
{
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
	if( (millis() - gPreviousL1) >= 1)
	{
		ControlHeadLight();
		
		gPreviousL1 = millis();
	}
	
	if( (millis() - gPreviousL5) >= 100)
	{
		//Motor drive control
		Motor_Control();
		
		#ifdef DSFOX
		ControlRunLED();
		#endif
		
		//Reset task
		gPreviousL5 = millis();
		
	}
}

/* Motor control Task (10Hz) */
void Motor_Control()
{
	uint16_t aSpeedRef = gSpeedCmd;
	uint16_t aPWMRef = 0;
	uint16_t aLPF_Ratio = gCV3_AccRatio;
	uint16_t aSpeedRefLinear = 0;
	uint32_t aSpeedRef_offseted = 0;
	
	//8bit LPF
	gPwmLPF_buf = ((gPwmLPF_buf * (256 - aLPF_Ratio)) >> 8) + ( aLPF_Ratio * aSpeedRef);

	//Linear Adjustment
	if( (gPwmLPF_buf >> 8) > 0)
	{
		//Accerelation
		aSpeedRef_offseted = (((gCV5_VMAX - (uint32_t)gCV2_Vstart) * (uint32_t)gPwmLPF_buf) >> 8) + ((uint32_t)gCV2_Vstart << 8);
		
		/* 16bitPWMでのリミッタ(後段は8bit) */
		if( (aSpeedRef_offseted >> 8) > gCV5_VMAX)
		{
			aSpeedRef_offseted = gCV5_VMAX << 8;
		}
		
		aSpeedRefLinear = (uint16_t)aSpeedRef_offseted;
		
		/*
		#ifdef _DEBUG
		Serial.print("PWM: ");
		Serial.println(aSpeedRef_offseted >> 8);
		#endif
		*/
	}
	else
	{
		aSpeedRefLinear = 0;
	}

	gPrevSpeed = gPwmLPF_buf;

	//Limiter (8bit PWM Output)
	aPWMRef = limitSpeed((uint16_t)(aSpeedRefLinear >> 8));
	
	//PWM出力
	if( aPWMRef == 0)
	{
		#if defined(__AVR_ATtiny85__)
		pwmWrite(motor_pwma, 0);
		pwmWrite(motor_pwmb, 0);
		#else
		analogWrite(MOTOR_PWM_A, 0);
		analogWrite(MOTOR_PWM_B, 0);
		#endif
	}
	else
	{
		//進行方向でPWMのABを切り替える
		#ifdef A4953
		if( gPwmDirv > 0){
			#if defined(__AVR_ATtiny85__)
			pwmWrite(motor_pwma, 255 - aPWMRef);
			pwmWrite(motor_pwmb, 255);
			#else
			analogWrite(MOTOR_PWM_A, 255 - aPWMRef);
			analogWrite(MOTOR_PWM_B, 255);
			#endif
		}
		else
		{
			#if defined(__AVR_ATtiny85__)
			pwmWrite(motor_pwma, 255);
			pwmWrite(motor_pwmb, 255 - aPWMRef);
			#else
			analogWrite(MOTOR_PWM_A, 255);
			analogWrite(MOTOR_PWM_B, 255 - aPWMRef);
			#endif
		}
		#else
		if( gPwmDirv > 0){
			#if defined(__AVR_ATtiny85__)
			pwmWrite(motor_pwma, aPWMRef);
			pwmWrite(motor_pwmb, 0);
			#else
			analogWrite(MOTOR_PWM_A, aPWMRef);
			analogWrite(MOTOR_PWM_B, 0);
			#endif
		}
		else
		{
			#if defined(__AVR_ATtiny85__)
			pwmWrite(motor_pwma, 0);
			pwmWrite(motor_pwmb, aPWMRef);
			#else
			analogWrite(MOTOR_PWM_A, 0);
			analogWrite(MOTOR_PWM_B, aPWMRef);
			#endif
		}
		#endif
	}
	
}

#if defined(__AVR_ATtiny85__)
//http://www.technoblogy.com/show?LE0
//void SetColour (int colour, int intensity)
void pwmWrite(int dir, int intensity)
{
  *Port[dir] = intensity;
}
#endif

uint16_t limitSpeed(uint16_t inSpeed)
{
	uint16_t aSpeed_ref = 0;
	
	if( inSpeed >= gCV5_VMAX)
	{
		aSpeed_ref = gCV5_VMAX;
	}
	else if( inSpeed <= 0)
	{
		aSpeed_ref = 0;
	}
/*	else if( (inSpeed > 0) && (inSpeed <= gCV2_Vstart))
	{
		aSpeed_ref = gCV2_Vstart;
	}*/
	else
	{
		aSpeed_ref = inSpeed;
	}
	
	return aSpeed_ref;
	
}

// 100ms interval Task
#ifdef DSFOX
void ControlRunLED()
{
	gRunCount++;
	if(gRunCount < 6)
	{
		if(gFuncBits[1] == 0)
		{
			digitalWrite(RUN_LED, 1);
		}
	}
	else
	{
		if(gFuncBits[1] == 1)
		{
			digitalWrite(RUN_LED, 0);
		}
		if(gRunCount > 10) gRunCount = 0;
	}
}
#endif

// 1ms interval Task
void ControlHeadLight(void)
{
	static uint8_t aFuncBits0 = 0;
	
	gBrightCount++;
	if (gBrightCount > 3)
	{
		gBrightCount = 0;
	}
	
	if(gFuncBits[0] == 0)
	{
		digitalWrite(PIN_FHEAD, 1);
		digitalWrite(PIN_RHEAD, 1);
		
		aFuncBits0 = 0;
		return;
	}
	
	if( gPwmDirv == 0) //FWD
	{
		// 進行方向を点灯
		digitalWrite(PIN_RHEAD, 1);
		
		// 停車中の減光
		if (gSpeedCmd <= 1)
		{
			if (gBrightCount < gCV33_Bright)
			{
				digitalWrite(PIN_FHEAD, 1); // Off
			}
			else
			{
				if(gCV19_Consist == 0) {
					digitalWrite(PIN_FHEAD, 0); // On
				} else {
					// 統括アドレスの時は点灯しない
					digitalWrite(PIN_FHEAD, 1); // Off
				}
			}
		}
		else
		{
			if(gCV19_Consist == 0)
			{
				digitalWrite(PIN_FHEAD, 0); // On
			}
			else // 統括アドレスの時は点灯しない
			{
				digitalWrite(PIN_FHEAD, 1); // Off
			}
		}
	}
	else //REV
	{
		// 後退方向を点灯
		digitalWrite(PIN_FHEAD, 1);
		
		// 停車中の減光
		if (gSpeedCmd <= 1)
		{
			if (gBrightCount < gCV33_Bright)
			{
				digitalWrite(PIN_RHEAD, 1); // Off
			}
			else
			{
				digitalWrite(PIN_RHEAD, 0); // On
			}
		}
		else
		{
			digitalWrite(PIN_RHEAD, 0); // On
		}
	}
	
	#ifdef _DEBUG
	if(aFuncBits0 == 0)
	{
		Serial.print("ControlHeadLight:");
		Serial.print("Dir=");
		Serial.print(gPwmDirv);
		Serial.print(",CV29dir=");
		Serial.print(gCV29Dir);
		Serial.print(",Consist=");
		Serial.println(gCV19_Consist);
		
		aFuncBits0 = 1;
	}
	#endif
}

//統括アドレス設定
extern void notifyDccConsist( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Cmd, uint8_t ConsistAddr)
{
	if(Addr != gCV1_SAddr) return;
	
	gCV19_Consist = ConsistAddr;
	gConsistDir = Cmd << 7;
	Dcc.setCV(CV_CONSIST, gConsistDir | gCV19_Consist);
	
	#ifdef _DEBUG
	Serial.print("notifyConsist:");
	Serial.print(gCV1_SAddr);
	Serial.print(",");
	Serial.print(gCV19_Consist);
	Serial.print(",");
	Serial.println(gConsistDir);
	#endif
}

//DCC速度信号の受信によるイベント
//extern void notifyDccSpeed( uint16_t Addr, uint8_t Speed, uint8_t ForwardDir, uint8_t MaxSpeed )
extern void notifyDccSpeed( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Speed, DCC_DIRECTION Dir, DCC_SPEED_STEPS SpeedSteps )
{
	if(gCV19_Consist == 0) {
		if(Addr != gCV1_SAddr) return;
	} else {
		if(Addr != gCV19_Consist) return;
	}
	
	uint16_t aSpeedRef = 0;
	
	//速度値の正規化(255を100%とする処理)
	if( Speed >= 1)
	{
		aSpeedRef = ((Speed - 1) * 255) / SpeedSteps;
	}
	else
	{
		aSpeedRef = 0;
	}
	
	//リミッタ
	if(aSpeedRef > gCV5_VMAX)
	{
		aSpeedRef = gCV5_VMAX;
	}
	
	if ( gCV29Dir > 0)
	{
		if(Dir == DCC_DIR_FWD) {
			gPwmDirv =128;
		} else {
			gPwmDirv =0;
		}
	}
	else
	{
		if(Dir == DCC_DIR_FWD) {
			gPwmDirv =0;
		} else {
			gPwmDirv =128;
		}
	}
	
	if(aSpeedRef != gSpeedCmd)
	{
		// デバッグメッセージ
		#ifdef _DEBUG
		Serial.print("notifyDccSpeed:");
		Serial.print(Addr);
		Serial.print(",");
		Serial.print(Dir);
		Serial.print(",");
		Serial.print(Speed);
		Serial.print(",");
		Serial.println(aSpeedRef);
		#endif
	}
	
	gSpeedCmd = aSpeedRef;
}

//ファンクション信号受信のイベント
//extern void notifyDccFunc( uint16_t Addr, FN_GROUP FuncGrp, uint8_t FuncState)
extern void notifyDccFunc( uint16_t Addr, DCC_ADDR_TYPE AddrType, FN_GROUP FuncGrp, uint8_t FuncState)
{
	if(gCV19_Consist == 0) {
		if(Addr != gCV1_SAddr) return;
	} else {
		if(Addr != gCV19_Consist) return;
	}
	
	switch(FuncGrp)
	{
	  case FN_0_4: // Function Group 1 F0 F4 F3 F2 F1
		if ((FuncState & FN_BIT_00) == 16) {
			if(gFuncBits[0] == 0) gFuncBits[0] = 1;
		} else {
			if(gFuncBits[0] == 1) gFuncBits[0] = 0;
		}
		if ((FuncState & FN_BIT_01) == 1) {
			if(gFuncBits[1] == 0) gFuncBits[1] = 1;
		} else {
			if(gFuncBits[1] == 1) gFuncBits[1] = 0;
		}
		if ((FuncState & FN_BIT_02) == 2) {
			if(gFuncBits[2] == 0) gFuncBits[2] = 1;
		} else {
			if(gFuncBits[2] == 1) gFuncBits[2] = 0;
		}
		if ((FuncState & FN_BIT_03) == 4) {
			if(gFuncBits[3] == 0) gFuncBits[3] = 1;
		} else {
			if(gFuncBits[3] == 1) gFuncBits[3] = 0;
		}
		if ((FuncState & FN_BIT_04) == 8) {
			if(gFuncBits[4] == 0) gFuncBits[4] = 1;
		} else {
			if(gFuncBits[4] == 1) gFuncBits[4] = 0;
		}
		break;
		
	  case FN_5_8: //Function Group 1 S FFFF == 1 F8 F7 F6 F5	&  == 0	 F12 F11 F10 F9 F8
		/*exec_function( 5, FunctionPin5, (FuncState & FN_BIT_05));
		exec_function( 6, FunctionPin6, (FuncState & FN_BIT_06)>>1 );
		exec_function( 7, FunctionPin7, (FuncState & FN_BIT_07)>>2 );
		exec_function( 8, FunctionPin8, (FuncState & FN_BIT_08)>>3 );
		*/
		break;
		
	  case FN_9_12:
		/*exec_function( 9, FunctionPin9,	(FuncState & FN_BIT_09));
		exec_function( 10, FunctionPin10, (FuncState & FN_BIT_10)>>1 );
		exec_function( 11, FunctionPin11, (FuncState & FN_BIT_11)>>2 );
		exec_function( 12, FunctionPin12, (FuncState & FN_BIT_12)>>3 );
		*/
		break;
		
	  case FN_13_20: //Function Group 2 FuncState == F20-F13 Function Control
		/*exec_function( 13, FunctionPin13, (FuncState & FN_BIT_13));
		exec_function( 14, FunctionPin14, (FuncState & FN_BIT_14)>>1 );
		exec_function( 15, FunctionPin15, (FuncState & FN_BIT_15)>>2 );
		exec_function( 16, FunctionPin16, (FuncState & FN_BIT_16)>>3 );
		*/
		break;
		
	  case FN_21_28:
		break;
	}
}
/*
void exec_function (int function, int pin, int FuncState)
{
	digitalWrite (pin, FuncState);
}
*/
